require("quadprog")
require("gtools")

#
#This function implements the algorithm for the Stage 1 of EDec.
#It takes as input the methylation profiles of complex samples,
#the set of loci with high variability in methylation across cell types,
#and the number of constituent cell types.
#It then estimates the methylation profiles of constituent cell types, and
#the proportions of constituent cell types in each input sample.
#
EDecStage1 = function(methMixtureSamples,cellTypeSpecificLoci,nCts,maxIts=2000,rssDiffStop=1e-10){
  
  if(sum(is.na(methMixtureSamples))>0){
    warning("Your input methylation profiles contain NA values. 
            Loci with NA values in any samples will not be included in the analysis, 
            and will not be present in cell type specific methylation profiles.")
  }
  methMixtureSamples = as.matrix(na.omit(methMixtureSamples))
  cellTypeSpecificLoci = intersect(cellTypeSpecificLoci,rownames(methMixtureSamples))
  mixturesOverCTSLoci = methMixtureSamples[cellTypeSpecificLoci,]
  nMixtures = ncol(methMixtureSamples)
  nLoci = nrow(mixturesOverCTSLoci)
  
  estProps = rdirichlet(nMixtures,1*rep(1/nCts,nCts))
  estMeth = estimateMethQP(mixturesOverCTSLoci,estProps)
  
  rss = norm(mixturesOverCTSLoci - estMeth%*%t(estProps),"F")^2
  it = 0
  rssDiff = Inf
  rssPerIteration = c()
  
  while(it < maxIts & rssDiff > rssDiffStop){
    rssPerIteration = c(rssPerIteration,rss)
    estProps = estimatePropsQP(mixturesOverCTSLoci,estMeth)
    estMeth = estimateMethQP(mixturesOverCTSLoci,estProps)
    newRss = norm(mixturesOverCTSLoci - estMeth%*%t(estProps),"F")^2
    rssDiff = rss - newRss
    rss = newRss
    it = it+1
  }
  
  nSamples = nMixtures*nLoci
  nParam = nCts*(nMixtures+nLoci)
  aic = nSamples*log(rss/nSamples)+ 2*nParam + (2*nParam*(nParam+1))/(nSamples-nParam-1)
  expVar = 1-(rss/(norm(mixturesOverCTSLoci,"F")^2))
  
  fullEstMeth = estimateMethQP(methMixtureSamples,estProps)
  
  result = list(fullEstMeth,estProps,it,expVar,rss,aic,rssPerIteration)
  names(result) = c("methylation","proportions","iterations","explained.variance","rss","aic","rss.per.iteration")
  return(result)
}

#
#This function will estimate the proportions of cell types assuming
#that the methylation profiles of constituent cell types are known
#It uses quadratic programming to find the best estimate
#
estimatePropsQP = function(methMixtureSamples, cellTypeSpecificMeth){  
  nCts = ncol(cellTypeSpecificMeth)
  nMixtures = ncol(methMixtureSamples)
  
  estimate = matrix(0, nMixtures, nCts)
  rownames(estimate) = colnames(methMixtureSamples)
  colnames(estimate) = colnames(cellTypeSpecificMeth)
  
  Amat = cbind(rep(-1,nCts), diag(nCts))
  b0vec = c(-1,rep(0,nCts))
  Dmat = t(cellTypeSpecificMeth)%*%cellTypeSpecificMeth
  
  deconv = function(x){
    dvec = t(cellTypeSpecificMeth)%*%x
    result = solve.QP(Dmat, dvec, Amat, b0vec, meq = 1)
    result$solution
  }
  estimate = t(apply(methMixtureSamples,2,deconv))
  return(estimate)
}

#
#This function will estimate the methylation profiles of constituent
#cell types, assuming that the proportions of each cell type in each
#sample are known. It uses the quadratic programming technique to do that.
#
estimateMethQP = function(methMixtureSamples, cellTypeProps){
  nCts = ncol(cellTypeProps)
  nLoci = nrow(methMixtureSamples)
  
  estimate = matrix(0,nrow=nrow(methMixtureSamples),ncol=ncol(cellTypeProps))
  rownames(estimate) = rownames(methMixtureSamples)
  colnames(estimate) = colnames(cellTypeProps)
  
  Amat = diag(rep(1, nCts))
  Amat = cbind(Amat, diag(rep(-1, nCts)))
  bvec = c(rep(0, nCts), rep(-1, nCts))
  Dmat = t(cellTypeProps) %*% cellTypeProps
  deconv = function(x){
    dvec = x %*% cellTypeProps
    result = solve.QP(Dmat, dvec, Amat, bvec=bvec, meq=0)
    result$solution
  }
  estimate = t(apply(methMixtureSamples,1,deconv))
  return(estimate)
}

#
#This function implements the second stage of the EDec method.
#It takes as input the gene expression profiles of complex samples,
#and the proportions of constituent cell types.
#It then estimates average and standard errors of cell type specific
#gene expression profiles.
#
EDecStage2 = function(geneExpMixtureSamples, cellTypeProps){
  geneExpMixtureSamples = as.matrix(geneExpMixtureSamples)
  
  nCts = ncol(cellTypeProps)
  nGenes = nrow(geneExpMixtureSamples)
  nSamples = nrow(cellTypeProps)
  
  estimate = matrix(0,nrow=nrow(geneExpMixtureSamples),ncol=ncol(cellTypeProps))
  rownames(estimate) = rownames(geneExpMixtureSamples)
  colnames(estimate) = colnames(cellTypeProps)
  
  Amat = diag(rep(1, nCts))
  bvec = rep(0, nCts)
  
  for(gene in 1:nGenes){
    Dmat = t(cellTypeProps) %*% cellTypeProps
    dvec = geneExpMixtureSamples[gene,] %*% cellTypeProps
    result = solve.QP(Dmat, dvec, Amat, bvec=bvec, meq=0)
    estimate[gene,] = result$solution
  }
  residuals = geneExpMixtureSamples-estimate%*%t(cellTypeProps)
  meanSquaredResiduals = apply(residuals^2,1,sum)/(nSamples-nCts)
  explainedVariances = 1-(apply(residuals^2,1,sum)/apply((geneExpMixtureSamples^2),1,sum))
  m = solve(t(cellTypeProps)%*%cellTypeProps)
  vars = matrix(0,nGenes,nCts)
  for(i in 1:nGenes){
    vars[i,] = meanSquaredResiduals[i]*diag(m)
  }
  sds = sqrt(vars)
  rownames(sds) = rownames(estimate)
  colnames(sds) = colnames(estimate)
  result = list(estimate,sds,nSamples-nCts,explainedVariances,residuals)
  names(result) = c("Means","Std.Errors","Degrees.of.freedom","Explained.Variances","Residuals")
  return(result)
}